import galsim
import contextlib
import numpy as np
import matplotlib.pyplot as plt
import ipywidgets
import batoid
from .builder import LSSTBuilder


def get_zk(opd):
    xs = np.linspace(-1, 1, opd.shape[0])
    ys = np.linspace(-1, 1, opd.shape[1])
    xs, ys = np.meshgrid(xs, ys)
    w = ~opd.mask
    basis = galsim.zernike.zernikeBasis(22, xs[w], ys[w], R_inner=0.61)
    zk, *_ = np.linalg.lstsq(basis.T, opd[w], rcond=None)
    return zk


def sub_ptt(opd):
    xs = np.linspace(-1, 1, opd.shape[0])
    ys = np.linspace(-1, 1, opd.shape[1])
    xs, ys = np.meshgrid(xs, ys)
    zk = get_zk(opd)
    opd -= galsim.zernike.Zernike(zk[:4], R_inner=0.61)(xs, ys)
    return opd


from matplotlib.colors import LinearSegmentedColormap
cmap = LinearSegmentedColormap.from_list('my_gradient', (
    # Edit this gradient at https://eltos.github.io/gradient/#0:99AEFF-33.3:0025B3-50:000000-66.7:C7030D-100:FF9FA3
    (0.000, (0.600, 0.682, 1.000)),
    (0.333, (0.000, 0.145, 0.702)),
    (0.500, (0.000, 0.000, 0.000)),
    (0.667, (0.780, 0.012, 0.051)),
    (1.000, (1.000, 0.624, 0.639))))


class VisualizeApp:
    def __init__(self, fiducial, wavelength, fea_dir="fea_legacy", bend_dir="bend_legacy", debug=None):
        self.nupdate = 0
        if debug is None:
            debug = contextlib.redirect_stdout(None)
        self.debug = debug
        self.fiducial = fiducial
        self.wavelength = wavelength
        self.fea_dir = fea_dir
        self.bend_dir = bend_dir

        # widget variables
        self.thx = 0.0
        self.thy = 0.0
        self.zenith = 0.0
        self.rotation = 0.0
        self.nphot = int(2e5)
        self.fwhm = 0.7

        self.do_M1M3_gravity = False
        self.do_M1M3_temperature = False
        self.M1M3_TBulk = 0.0
        self.M1M3_TxGrad = 0.0
        self.M1M3_TyGrad = 0.0
        self.M1M3_TzGrad = 0.0
        self.M1M3_TrGrad = 0.0
        self.do_M1M3_LUT = False
        self.M1M3_LUT_error = 0.0
        self.M1M3_LUT_seed = 0

        self.do_M2_gravity = False
        self.do_M2_temperature = False
        self.M2_TzGrad = 0.0
        self.M2_TrGrad = 0.0

        self.do_camera_gravity = False
        self.do_camera_temperature = False
        self.camera_TBulk = 0.0

        self.dof_idx = 0
        self.dof_value = 0.0

        self.actuator_idx = 0
        self.actuator_value = 0.0

        # Output
        self.view = self._view()

        # Controls
        kwargs = {'layout':{'width':'180px'}}
        self.thx_control = ipywidgets.FloatText(value=0.0, step=0.25, description='Field x (deg)', **kwargs)
        self.thy_control = ipywidgets.FloatText(value=0.0, step=0.25, description='Field y (deg)', **kwargs)
        self.zenith_control = ipywidgets.BoundedFloatText(value=0.0, step=1.0, min=0.0, max=90.0, description="Zenith (deg)", **kwargs)
        self.rotation_control = ipywidgets.BoundedFloatText(value=0.0, step=1.0, min=-90.0, max=90.0, description="Rotation (deg)", **kwargs)
        self.nphot_control = ipywidgets.IntText(value=int(2e5), description='# phot')
        self.fwhm_control = ipywidgets.FloatText(value=0.7, description='seeing (arcsec)')
        self.point_control = ipywidgets.VBox([
            self.thx_control,
            self.thy_control,
            self.zenith_control,
            self.rotation_control,
            self.nphot_control,
            self.fwhm_control
        ])

        self.M1M3_gravity_checkbox = ipywidgets.Checkbox(value=False, description="Gravity")
        self.M1M3_temperature_checkbox = ipywidgets.Checkbox(value=False, description="Temperature")
        self.M1M3_TBulk_control = ipywidgets.FloatText(value=0.0, step=0.1, description="TBulk")
        self.M1M3_TxGrad_control = ipywidgets.FloatText(value=0.0, step=0.01, description="TxGrad")
        self.M1M3_TyGrad_control = ipywidgets.FloatText(value=0.0, step=0.01, description="TyGrad")
        self.M1M3_TzGrad_control = ipywidgets.FloatText(value=0.0, step=0.01, description="TzGrad")
        self.M1M3_TrGrad_control = ipywidgets.FloatText(value=0.0, step=0.01, description="TrGrad")
        self.M1M3_LUT_checkbox = ipywidgets.Checkbox(value=False, description="Apply LUT")
        self.M1M3_LUT_error_control = ipywidgets.FloatText(value=0.0, step=0.001, description="force error")
        self.M1M3_LUT_seed_control = ipywidgets.BoundedIntText(value=0, description="error seed", min=0)
        self.M1M3_control = ipywidgets.VBox([
            self.M1M3_gravity_checkbox,
            self.M1M3_temperature_checkbox,
            self.M1M3_TBulk_control,
            self.M1M3_TxGrad_control,
            self.M1M3_TyGrad_control,
            self.M1M3_TzGrad_control,
            self.M1M3_TrGrad_control,
            self.M1M3_LUT_checkbox,
            self.M1M3_LUT_error_control,
            self.M1M3_LUT_seed_control
        ])
        self.M2_gravity_checkbox = ipywidgets.Checkbox(value=False, description="Gravity")
        self.M2_temperature_checkbox = ipywidgets.Checkbox(value=False, description="Temperature")
        self.M2_TzGrad_control = ipywidgets.FloatText(value=0.0, step=0.01, description="TzGrad")
        self.M2_TrGrad_control = ipywidgets.FloatText(value=0.0, step=0.01, description="TrGrad")
        self.M2_control = ipywidgets.VBox([
            self.M2_gravity_checkbox,
            self.M2_temperature_checkbox,
            self.M2_TzGrad_control,
            self.M2_TrGrad_control
        ])

        self.camera_gravity_checkbox = ipywidgets.Checkbox(value=False, description="Gravity")
        self.camera_temperature_checkbox = ipywidgets.Checkbox(value=False, description="Temperature")
        self.camera_TBulk_control = ipywidgets.BoundedFloatText(value=0.0, step=0.1, description="TBulk", min=-10.0, max=25.0)
        self.camera_control = ipywidgets.VBox([
            self.camera_gravity_checkbox,
            self.camera_temperature_checkbox,
            self.camera_TBulk_control
        ])

        self.dof_idx_control = ipywidgets.BoundedIntText(value=0, min=0, max=49, description="index")
        self.dof_value_control = ipywidgets.FloatText(value=0.0, step=0.1, description="value (micron)")

        self.dof_control = ipywidgets.VBox([
            self.dof_idx_control,
            self.dof_value_control
        ])

        self.actuator_idx_control = ipywidgets.BoundedIntText(value=0, min=0, max=255, description="index")
        self.actuator_value_control = ipywidgets.FloatText(value=0.0, step=10.0, description="value (Newton)")


        self.actuator_control = ipywidgets.VBox([
            self.actuator_idx_control,
            self.actuator_value_control
        ])

        self.tab_control = ipywidgets.Tab()
        self.tab_control.children = [
            self.point_control,
            self.M1M3_control,
            self.M2_control,
            self.camera_control,
            self.dof_control,
            self.actuator_control,
        ]

        # Lame interface, but works
        self.tab_control.set_title(0, 'Point')
        self.tab_control.set_title(1, 'M1M3')
        self.tab_control.set_title(2, 'M2')
        self.tab_control.set_title(3, 'Cam')
        self.tab_control.set_title(4, 'DOF')
        self.tab_control.set_title(5, 'Force')
        self.tab_control.layout.width='340px'

        self.thx_control.observe(self.handle_thx, 'value')
        self.thy_control.observe(self.handle_thy, 'value')
        self.zenith_control.observe(self.handle_zenith, 'value')
        self.rotation_control.observe(self.handle_rotation, 'value')
        self.nphot_control.observe(self.handle_nphot, 'value')
        self.fwhm_control.observe(self.handle_fwhm, 'value')

        self.M1M3_gravity_checkbox.observe(self.handle_M1M3_grav_check, 'value')
        self.M1M3_temperature_checkbox.observe(self.handle_M1M3_temp_check, 'value')
        self.M1M3_TBulk_control.observe(self.handle_M1M3_TBulk, 'value')
        self.M1M3_TxGrad_control.observe(self.handle_M1M3_TxGrad, 'value')
        self.M1M3_TyGrad_control.observe(self.handle_M1M3_TyGrad, 'value')
        self.M1M3_TzGrad_control.observe(self.handle_M1M3_TzGrad, 'value')
        self.M1M3_TrGrad_control.observe(self.handle_M1M3_TrGrad, 'value')

        self.M1M3_LUT_checkbox.observe(self.handle_M1M3_LUT_check, 'value')
        self.M1M3_LUT_error_control.observe(self.handle_M1M3_LUT_error, 'value')
        self.M1M3_LUT_seed_control.observe(self.handle_M1M3_LUT_seed, 'value')

        self.M2_gravity_checkbox.observe(self.handle_M2_grav_check, 'value')
        self.M2_temperature_checkbox.observe(self.handle_M2_temp_check, 'value')
        self.M2_TzGrad_control.observe(self.handle_M2_TzGrad, 'value')
        self.M2_TrGrad_control.observe(self.handle_M2_TrGrad, 'value')

        self.camera_gravity_checkbox.observe(self.handle_camera_grav_check, 'value')
        self.camera_temperature_checkbox.observe(self.handle_camera_temp_check, 'value')
        self.camera_TBulk_control.observe(self.handle_camera_TBulk, 'value')

        self.dof_idx_control.observe(self.handle_dof_idx, 'value')
        self.dof_value_control.observe(self.handle_dof_value, 'value')

        self.actuator_idx_control.observe(self.handle_actuator_idx, 'value')
        self.actuator_value_control.observe(self.handle_actuator_value, 'value')

        self.update()

    def handle_thx(self, change):
        self.thx = change['new']
        self.update()

    def handle_thy(self, change):
        self.thy = change['new']
        self.update()

    def handle_zenith(self, change):
        self.zenith = change['new']
        self.update()

    def handle_rotation(self, change):
        self.rotation = change['new']
        self.update()

    def handle_nphot(self, change):
        self.nphot = change['new']
        self.update()

    def handle_fwhm(self, change):
        self.fwhm = change['new']
        self.update()

    def handle_M1M3_grav_check(self, change):
        self.do_M1M3_gravity = not self.do_M1M3_gravity
        self.update()

    def handle_M1M3_temp_check(self, change):
        self.do_M1M3_temperature = not self.do_M1M3_temperature
        self.update()

    def handle_M1M3_TBulk(self, change):
        self.M1M3_TBulk = change['new']
        self.update()

    def handle_M1M3_TxGrad(self, change):
        self.M1M3_TxGrad = change['new']
        self.update()

    def handle_M1M3_TyGrad(self, change):
        self.M1M3_TyGrad = change['new']
        self.update()

    def handle_M1M3_TzGrad(self, change):
        self.M1M3_TzGrad = change['new']
        self.update()

    def handle_M1M3_TrGrad(self, change):
        self.M1M3_TrGrad = change['new']
        self.update()

    def handle_M1M3_LUT_check(self, change):
        self.do_M1M3_LUT = not self.do_M1M3_LUT
        self.update()

    def handle_M1M3_LUT_error(self, change):
        self.M1M3_LUT_error = change['new']
        self.update()

    def handle_M1M3_LUT_seed(self, change):
        self.M1M3_LUT_seed = change['new']
        self.update()

    def handle_M2_grav_check(self, change):
        self.do_M2_gravity = not self.do_M2_gravity
        self.update()

    def handle_M2_temp_check(self, change):
        self.do_M2_temperature = not self.do_M2_temperature
        self.update()

    def handle_M2_TzGrad(self, change):
        self.M2_TzGrad = change['new']
        self.update()

    def handle_M2_TrGrad(self, change):
        self.M2_TrGrad = change['new']
        self.update()

    def handle_camera_grav_check(self, change):
        self.do_camera_gravity = not self.do_camera_gravity
        self.update()

    def handle_camera_temp_check(self, change):
        self.do_camera_temperature = not self.do_camera_temperature
        self.update()

    def handle_camera_TBulk(self, change):
        self.camera_TBulk = change['new']
        self.update()

    def handle_dof_idx(self, change):
        self.dof_idx = change['new']
        self.update()

    def handle_dof_value(self, change):
        self.dof_value = change['new']
        self.update()

    def handle_actuator_idx(self, change):
        self.actuator_idx = change['new']
        self.update()

    def handle_actuator_value(self, change):
        self.actuator_value = change['new']
        self.update()

    def _view(self):
        self._fig, axes = plt.subplots(
            nrows=2, ncols=3,
            figsize=(6, 6),
            dpi=100,
            facecolor='#111111'
        )
        self._canvas = self._fig.canvas
        self._canvas.header_visible=False

        # Top Left/Center = wavefronts
        self.wfp_ax = axes[0,0]
        self.wfp_img = self.wfp_ax.imshow(
            np.ones((255, 255)),
            vmin=-1.0, vmax=1.0,
            cmap='seismic',
            extent=np.r_[1, -1, 1, -1]*4.18
        )
        self.wfp_ax.set_title("Wavefront", color='w')

        self.wfr_ax = axes[0,1]
        self.wfr_img = self.wfr_ax.imshow(
            np.zeros((255, 255)),
            vmin=-1.0, vmax=1.0,
            cmap='seismic',
            extent=np.r_[-1, 1, -1, 1]*4.18
        )
        self.wfr_ax.set_title("Wavefront residual", color='w')

        # Bottom Left/Center = donuts
        self.intra_ax = axes[1,0]
        self.intra_img = self.intra_ax.imshow(
            np.zeros((255, 255)),
            vmin=0.0, vmax=1.0,
            cmap='plasma',
            extent=np.r_[-1, 1, -1, 1]*0.2*255/2
        )
        self.intra_ax.set_title("Intra donut", color='w')

        self.extra_ax = axes[1,1]
        self.extra_img = self.extra_ax.imshow(
            np.zeros((255, 255)),
            vmin=0.0, vmax=1.0,
            cmap='plasma',
            extent=np.r_[-1, 1, -1, 1]*0.2*255/2
        )
        self.extra_ax.set_title("Extra donut", color='w')

        # Top Right = M1M3
        self.m1m3_ax = axes[0,2]
        self.m1m3_ax.set_facecolor('#111111')
        x = np.linspace(-4.18, 4.18, 255)
        xx, yy = np.meshgrid(x, x)
        rr = np.hypot(xx, yy)
        self.m1m3_mask = (
            rr < 0.55
        ) | (
            (rr > 2.508) & (rr < 2.558)
        ) | (
            rr > 4.18
        )
        m1m3_arr = np.ma.array(
            np.zeros((255, 255)),
            mask=self.m1m3_mask
        )
        self.m1m3_img = self.m1m3_ax.imshow(
            m1m3_arr,
            vmin=-1.0, vmax=1.0,
            cmap=cmap,
            extent=np.r_[1, -1, 1, -1]*4.18
        )
        self.m1m3_ax.set_title("M1M3", color='w')

        # Bottom Right = M2
        self.m2_ax = axes[1,2]
        self.m2_ax.set_facecolor('#111111')
        x = np.linspace(-1.71, 1.71, 255)
        xx, yy = np.meshgrid(x, x)
        rr = np.hypot(xx, yy)
        self.m2_mask = (rr < 0.9) | (rr > 1.71)
        m2_arr = np.ma.array(
            np.zeros((255, 255)),
            mask=self.m2_mask
        )
        self.m2_img = self.m2_ax.imshow(
            m2_arr,
            vmin=-1.0, vmax=1.0,
            cmap=cmap,
            extent=np.r_[1, -1, 1, -1]*1.71
        )
        self.m2_ax.set_title("M2", color='w')

        self._fig.tight_layout()
        out = ipywidgets.Output()
        with out:
            plt.show(self._fig)
        return out

    def update(self):
        self.nupdate += 1
        # Get the telescopes
        builder = LSSTBuilder(self.fiducial, self.fea_dir, self.bend_dir)
        tel0 = builder.build()
        tel0 = tel0.withLocallyRotatedOptic(
            "LSSTCamera", batoid.RotZ(np.deg2rad(self.rotation))
        )

        if self.do_M1M3_gravity:
            builder = builder.with_m1m3_gravity(
                np.deg2rad(self.zenith)
            )
        if self.do_M1M3_temperature:
            builder = builder.with_m1m3_temperature(
                self.M1M3_TBulk,
                self.M1M3_TxGrad,
                self.M1M3_TyGrad,
                self.M1M3_TzGrad,
                self.M1M3_TrGrad
            )
        if self.do_M1M3_LUT:
            builder = builder.with_m1m3_lut(
                np.deg2rad(self.zenith),
                self.M1M3_LUT_error,
                self.M1M3_LUT_seed
            )

        if self.do_M2_gravity:
            builder = builder.with_m2_gravity(
                np.deg2rad(self.zenith)
            )
        if self.do_M2_temperature:
            builder = builder.with_m2_temperature(
                self.M2_TzGrad, self.M2_TrGrad
            )

        if self.do_camera_gravity:
            builder = builder.with_camera_gravity(
                np.deg2rad(self.zenith),
                np.deg2rad(self.rotation)
            )
        if self.do_camera_temperature:
            builder = builder.with_camera_temperature(
                self.camera_TBulk
            )

        if self.dof_value != 0.0:
            dof = np.zeros(50)
            dof[self.dof_idx] = self.dof_value
            builder = builder.with_aos_dof(dof)

        if self.actuator_value != 0.0:
            forces = np.zeros(256)
            forces[self.actuator_idx] = self.actuator_value
            builder = builder.with_m1m3_extra_forces(forces)

        tel1 = builder.build()
        tel1 = tel1.withLocallyRotatedOptic(
            "LSSTCamera", batoid.RotZ(np.deg2rad(self.rotation))
        )

        # Update wavefront
        wf0 = sub_ptt(batoid.wavefront(
            tel0,
            np.deg2rad(self.thx), np.deg2rad(self.thy),
            self.wavelength,
            nx=255
        ).array)
        wf1 = sub_ptt(batoid.wavefront(
            tel1,
            np.deg2rad(self.thx), np.deg2rad(self.thy),
            self.wavelength,
            nx=255
        ).array)
        self.wfr_img.set_array(wf1-wf0)
        self.wfp_img.set_array(wf1)

        # Make some donuts
        intra = tel1.withGloballyShiftedOptic("Detector", (0, 0, -0.0015))
        extra = tel1.withGloballyShiftedOptic("Detector", (0, 0, +0.0015))
        rv = batoid.RayVector.asPolar(
            optic=tel1, wavelength=self.wavelength,
            nrandom=self.nphot,
            theta_x=np.deg2rad(self.thx),
            theta_y=np.deg2rad(self.thy)
        )
        rvi = intra.trace(rv.copy())
        rve = extra.trace(rv)

        rvi.x[:] -= np.mean(rvi.x[~rvi.vignetted])
        rvi.y[:] -= np.mean(rvi.y[~rvi.vignetted])
        rve.x[:] -= np.mean(rve.x[~rve.vignetted])
        rve.y[:] -= np.mean(rve.y[~rve.vignetted])

        # Convolve in a Gaussian
        scale = 10e-6 * self.fwhm/2.35/0.2
        rvi.x[:] += np.random.normal(scale=scale, size=len(rvi))
        rvi.y[:] += np.random.normal(scale=scale, size=len(rvi))
        rve.x[:] += np.random.normal(scale=scale, size=len(rve))
        rve.y[:] += np.random.normal(scale=scale, size=len(rve))

        # Bin rays
        di, _, _ = np.histogram2d(
            rvi.y[~rvi.vignetted], rvi.x[~rvi.vignetted], bins=255,
            range=[[-127*10e-6, 127*10e-6], [-127*10e-6, 127*10e-6]]
        )
        de, _, _ = np.histogram2d(
            rve.y[~rve.vignetted], rve.x[~rve.vignetted], bins=255,
            range=[[-127*10e-6, 127*10e-6], [-127*10e-6, 127*10e-6]]
        )

        self.intra_img.set_array(di/np.max(di))
        self.extra_img.set_array(de/np.max(de))

        # Update M1M3 shape
        x = np.linspace(-4.18, 4.18, 255)
        xx, yy = np.meshgrid(x, x)
        rr = np.hypot(xx, yy)
        w1 = (rr >= 2.558) & (rr <= 4.18)
        w3 = (rr >= 0.55) & (rr <= 2.508)
        sag = np.zeros((255, 255))
        sag[w1] = tel1['M1'].surface.sag(xx[w1], yy[w1])
        sag[w1] -= tel0['M1'].surface.sag(xx[w1], yy[w1])
        sag[w3] = tel1['M3'].surface.sag(xx[w3], yy[w3])
        sag[w3] -= tel0['M3'].surface.sag(xx[w3], yy[w3])
        sag *= 1e6
        m1m3_arr = np.ma.array(sag, mask=self.m1m3_mask)
        self.m1m3_img.set_array(m1m3_arr)

        # Update M2 shape
        x = np.linspace(-1.71, 1.71, 255)
        xx, yy = np.meshgrid(x, x)
        rr = np.hypot(xx, yy)
        w = (rr >= 0.9) & (rr <= 1.71)
        sag = np.zeros((255, 255))
        sag[w] = tel1['M2'].surface.sag(xx[w], yy[w])
        sag[w] -= tel0['M2'].surface.sag(xx[w], yy[w])
        sag *= 1e6
        m2_arr = np.ma.masked_array(sag, mask=self.m2_mask)
        self.m2_img.set_array(m2_arr)

        self._canvas.draw()

    def display(self):
        from IPython.display import display
        self.app = ipywidgets.HBox([
            self.view,
            self.tab_control
        ])

        display(self.app)
        self.update()  # not sure why this is needed, but works with this here.



__version__ = "0.4"
tmp = __version__
if "rc" in tmp:  # pragma: no cover
    tmp = tmp[:tmp.find("rc")]
__version_info__ = tuple(map(int, tmp.split('.')))

from .builder import *
from .visualize import *
from .align_game import AlignGame

from pathlib import Path
datadir = Path(__file__).parent / "data"



from pathlib import Path

import batoid
import batoid_rubin
import galsim
import numpy as np


fea_dir = Path(batoid_rubin.datadir) / "fea_legacy"
bend_dir = Path(batoid_rubin.datadir) / "bend"

zen = 30 * galsim.degrees
rot = 15 * galsim.degrees


def test_builder():
    fiducial = batoid.Optic.fromYaml("LSST_r.yaml")
    builder = batoid_rubin.builder.LSSTBuilder(fiducial, fea_dir, bend_dir)
    builder = (
        builder
        .with_m1m3_gravity(zen)
        .with_m1m3_temperature(0.0, 0.1, -0.1, 0.1, 0.1)
        .with_m2_gravity(zen)
        .with_m2_temperature(0.1, 0.1)
        .with_aos_dof(np.array([0]*19+[1]+[0]*30))
        .with_m1m3_lut(zen, 0.0, 0)
        .with_extra_zk([0]*4+[1e-9], 0.61)
    )

    telescope = builder.build()

    # Check that default dirs work
    builder2 = batoid_rubin.LSSTBuilder(fiducial)
    builder2 = (
        builder2
        .with_m1m3_gravity(zen)
        .with_m1m3_temperature(0.0, 0.1, -0.1, 0.1, 0.1)
        .with_m2_gravity(zen)
        .with_m2_temperature(0.1, 0.1)
        .with_aos_dof(np.array([0]*19+[1]+[0]*30))
        .with_m1m3_lut(zen, 0.0, 0)
        .with_extra_zk([0]*4+[1e-9], 0.61)
    )
    telescope2 = builder2.build()
    assert telescope == telescope2

    # Check float interface too.
    builder3 = batoid_rubin.LSSTBuilder(fiducial)
    builder3 = (
        builder3
        .with_m1m3_gravity(zen.rad)
        .with_m1m3_temperature(0.0, 0.1, -0.1, 0.1, 0.1)
        .with_m2_gravity(zen.rad)
        .with_m2_temperature(0.1, 0.1)
        .with_aos_dof(np.array([0]*19+[1]+[0]*30))
        .with_m1m3_lut(zen.rad, 0.0, 0)
        .with_extra_zk([0]*4+[1e-9], 0.61)
    )
    telescope3 = builder3.build()
    assert telescope == telescope3


def test_attr():
    builder = batoid_rubin.LSSTBuilder(batoid.Optic.fromYaml("LSST_r.yaml"))
    assert hasattr(builder.with_m1m3_gravity, "_req_params")


def test_ep_phase():
    fiducial = batoid.Optic.fromYaml("LSST_r.yaml")
    builder = batoid_rubin.builder.LSSTBuilder(fiducial, fea_dir, bend_dir)
    builder = (
        builder
        .with_m1m3_gravity(zen)
        .with_m1m3_temperature(0.0, 0.1, -0.1, 0.1, 0.1)
        .with_m2_gravity(zen)
        .with_m2_temperature(0.1, 0.1)
        .with_aos_dof(np.array([0]*19+[1]+[0]*30))
        .with_m1m3_lut(zen, 0.0, 0)
        .with_extra_zk([0]*4+[1e-9], 0.61)
    )
    telescope = builder.build()
    thx = 0.01
    thy = 0.01
    wavelength=622e-9
    zk = batoid.zernike(
        telescope, thx, thy, wavelength,
        nx=128, jmax=28, eps=0.61
    )
    # Now try to zero-out the wavefront

    builder1 = builder.with_extra_zk(
        zk*wavelength, 0.61
    )
    telescope1 = builder1.build()
    zk1 = batoid.zernike(
        telescope1, thx, thy, wavelength,
        nx=128, jmax=28, eps=0.61
    )

    np.testing.assert_allclose(
        zk1[4:], 0.0, atol=2e-3
    )  # 0.002 waves isn't so bad


def test_modes_permutation():
    """Test that permuting both dof and use_m1m3_modes identically gives same
    result as no permutation.
    """
    fiducial = batoid.Optic.fromYaml("LSST_r.yaml")
    builder1 = batoid_rubin.builder.LSSTBuilder(
        fiducial,
        use_m1m3_modes=list(range(20)),
        use_m2_modes=list(range(20))
    )
    rays = batoid.RayVector.asPolar(
        optic=fiducial,
        wavelength=622e-9,
        theta_x=0.01,
        theta_y=0.01,
        nrad=10,
        naz=60,
    )

    rng = np.random.default_rng(57721)
    for _ in range(10):
        p1 = rng.permutation(20)
        p2 = rng.permutation(20)
        builder2 = batoid_rubin.builder.LSSTBuilder(
            fiducial,
            use_m1m3_modes=p1,
            use_m2_modes=p2
        )
        rigid_dof = np.zeros(10)
        m1m3_dof = rng.uniform(-1e-6, 1e-6, size=20)
        m2_dof = rng.uniform(-1e-6, 1e-6, size=20)
        dof1 = np.concatenate([rigid_dof, m1m3_dof, m2_dof])
        dof2 = np.concatenate([rigid_dof, m1m3_dof[p1], m2_dof[p2]])
        scope1 = builder1.with_aos_dof(dof1).build()
        scope2 = builder2.with_aos_dof(dof2).build()

        trays1 = scope1.trace(rays.copy())
        trays2 = scope2.trace(rays.copy())

        np.testing.assert_allclose(
            trays1.r, trays2.r, rtol=0, atol=1e-15
        )
        np.testing.assert_allclose(
            trays1.v, trays2.v, rtol=0, atol=1e-15
        )
        np.testing.assert_equal(
            trays1.vignetted, trays2.vignetted
        )
        np.testing.assert_equal(
            trays1.failed, trays2.failed
        )

def test_subsys_dof():
    rng = np.random.default_rng(5772156649)
    fiducial = batoid.Optic.fromYaml("LSST_r.yaml")

    # Use a subset of the permuted modes to maximally stress the code.
    use_m1m3_modes = rng.permutation(20)[:18]
    use_m2_modes = rng.permutation(20)[:16]
    builder = batoid_rubin.builder.LSSTBuilder(
        fiducial,
        use_m1m3_modes=use_m1m3_modes,
        use_m2_modes=use_m2_modes
    )
    for _ in range(10):
        m2_dz = rng.uniform(-1, 1)
        m2_dx = rng.uniform(-100, 100)
        m2_dy = rng.uniform(-100, 100)
        m2_rx = rng.uniform(-1, 1)
        m2_ry = rng.uniform(-1, 1)

        cam_dz = rng.uniform(-1, 1)
        cam_dx = rng.uniform(-100, 100)
        cam_dy = rng.uniform(-100, 100)
        cam_rx = rng.uniform(-1, 1)
        cam_ry = rng.uniform(-1, 1)

        m1m3_bend = rng.uniform(-0.05, 0.05, size=18)
        m2_bend = rng.uniform(-0.05, 0.05, size=16)

        m2_dof = [m2_dz, m2_dx, m2_dy, m2_rx, m2_ry]
        cam_dof = [cam_dz, cam_dx, cam_dy, cam_rx, cam_ry]
        dof = np.concatenate([
            m2_dof,
            cam_dof,
            m1m3_bend,
            m2_bend
        ])

        builder1 = builder.with_aos_dof(dof)
        builder2 = builder.with_m2_rigid(
            dz=m2_dz, dx=m2_dx, dy=m2_dy,
            rx=m2_rx*galsim.arcsec, ry=m2_ry*galsim.arcsec
        ).with_camera_rigid(
            dz=cam_dz, dx=cam_dx, dy=cam_dy,
            rx=cam_rx*galsim.arcsec, ry=cam_ry*galsim.arcsec
        ).with_m1m3_bend(m1m3_bend).with_m2_bend(m2_bend)
        builder3 = builder.with_m2_rigid(
            dof=m2_dof
        ).with_camera_rigid(
            dof=cam_dof
        ).with_m1m3_bend(m1m3_bend).with_m2_bend(m2_bend)

        scope1 = builder1.build()
        scope2 = builder2.build()
        scope3 = builder3.build()

        assert scope1 == scope2
        assert scope1 == scope3


if __name__ == "__main__":
    test_builder()
    test_attr()
    test_ep_phase()
    test_modes_permutation()
    test_subsys_dof()



from pathlib import Path
from functools import lru_cache

import astropy.io.fits as fits
import numpy as np
from scipy.interpolate import CloughTocher2DInterpolator


def _node_to_grid(nodex, nodey, nodez, grid_coords):
    """Convert FEA nodes positions into grid of z displacements,
    first derivatives, and mixed 2nd derivative.

    Parameters
    ----------
    nodex, nodey, nodez : ndarray (M, )
        Positions of nodes
    grid_coords : ndarray (2, N)
        Output grid positions in x and y

    Returns
    -------
    grid : ndarray (4, N, N)
        1st slice is interpolated z-position.
        2nd slice is interpolated dz/dx
        3rd slice is interpolated dz/dy
        4th slice is interpolated d2z/dxdy
    """
    interp = CloughTocher2DInterpolator(
        np.array([nodex, nodey]).T,
        nodez,
        fill_value=0.0
    )

    x, y = grid_coords
    nx = len(x)
    ny = len(y)
    out = np.zeros((4, ny, nx))
    # Approximate derivatives with finite differences.  Make the finite
    # difference spacing equal to 1/10th the grid spacing.
    dx = np.mean(np.diff(x))*1e-1
    dy = np.mean(np.diff(y))*1e-1
    x, y = np.meshgrid(x, y)
    out[0] = interp(x, y)
    out[1] = (interp(x+dx, y) - interp(x-dx, y))/(2*dx)
    out[2] = (interp(x, y+dy) - interp(x, y-dy))/(2*dy)
    out[3] = (
        interp(x+dx, y+dy) -
        interp(x-dx, y+dy) -
        interp(x+dx, y-dy) +
        interp(x-dx, y-dy)
    )/(4*dx*dy)

    # Zero out the central hole
    r = np.hypot(x, y)
    rmin = np.min(np.hypot(nodex, nodey))
    w = r < rmin
    out[:, w] = 0.0

    return out


@lru_cache(maxsize=100)
def _fits_cache(datadir, fn):
    """Cache loading fits file data table

    Parameters
    ----------
    datadir : str
        Directory containing the fits file
    fn : string
        File name from datadir to load and cache

    Returns
    -------
    out : ndarray
        Loaded data.
    """
    return fits.getdata(Path(datadir) / fn)


def attach_attr(**kwargs):
    """Decorator for attaching attributes to functions
    """
    def inner(f):
        for k, v in kwargs.items():
            setattr(f, k, v)
        return f
    return inner



import contextlib
from functools import cached_property
from pathlib import Path
import yaml

import batoid
import batoid_rubin
import danish
import ipywidgets
import matplotlib.pyplot as plt
import numpy as np
from scipy.linalg import lstsq
from scipy.optimize import least_squares


class Raft:
    def __init__(self, name, thx, thy, nphot, img, fiducial, wavelength=500e-9):
        self.name = name
        self.thx = thx
        self.thy = thy
        self.nphot = nphot
        self.img = img
        self.fiducial = fiducial
        self.wavelength = wavelength

        self.bins = img.get_array().shape[0]
        bo2 = self.bins//2
        self.range = [[-bo2*10e-6, bo2*10e-6], [-bo2*10e-6, bo2*10e-6]]

        if "in" in self.name:
            telescope = self.fiducial.withGloballyShiftedOptic(
                "Detector", [0, 0, -1.5e-3]
            )
        elif "ex" in self.name:
            telescope = self.fiducial.withGloballyShiftedOptic(
                "Detector", [0, 0, 1.5e-3]
            )
        else:
            telescope = self.fiducial
        self.z_ref = batoid.zernikeTA(
            telescope, np.deg2rad(self.thx), np.deg2rad(self.thy),
            self.wavelength,
            nrad=20, naz=120,
            reference='chief',
            jmax=66, eps=0.61
        )*self.wavelength  # meters

    def draw(self, telescope, seeing, rng=None):
        if rng is None:
            rng = np.random.default_rng()
        elif isinstance(rng, int):
            rng = np.random.default_rng(rng)
        if "in" in self.name:
            telescope = telescope.withGloballyShiftedOptic(
                "Detector", [0, 0, -1.5e-3]
            )
        elif "ex" in self.name:
            telescope = telescope.withGloballyShiftedOptic(
                "Detector", [0, 0, 1.5e-3]
            )
        rv = batoid.RayVector.asPolar(
            optic=telescope, wavelength=self.wavelength,
            nrandom=self.nphot, rng=rng,
            theta_x=np.deg2rad(self.thx), theta_y=np.deg2rad(self.thy)
        )
        telescope.trace(rv)

        rv.x[:] -= np.mean(rv.x[~rv.vignetted])
        rv.y[:] -= np.mean(rv.y[~rv.vignetted])

        # Convolve in a Gaussian
        scale = 10e-6 * seeing/2.35/0.2
        rv.x[:] += rng.normal(scale=scale, size=len(rv))
        rv.y[:] += rng.normal(scale=scale, size=len(rv))

        # Bin rays
        psf, _, _ = np.histogram2d(
            rv.y[~rv.vignetted], rv.x[~rv.vignetted], bins=self.bins,
            range=self.range
        )
        self.img.set_array(psf/np.max(psf))


class AlignGame:
    def __init__(self, debug=None, control_log=None, rng=None, nthread=8):
        batoid._batoid.set_nthreads(nthread)

        if debug is None:
            debug = contextlib.redirect_stdout(None)
        self.debug = debug
        if control_log is None:
            control_log = contextlib.redirect_stdout(None)
        self.control_log = control_log
        if rng is None:
            rng = np.random.default_rng()
        elif isinstance(rng, int):
            rng = np.random.default_rng(rng)
        self.rng = rng
        self.offset_rng = np.random.default_rng(rng.integers(2**63))

        self.fiducial = batoid.Optic.fromYaml("LSST_r.yaml")
        self.builder = batoid_rubin.LSSTBuilder(
            self.fiducial,
            fea_dir="fea_legacy",
            bend_dir="bend_legacy"
        )
        self.wavelength = 500e-9

        # widget variables
        self.m2_dz = 0.0
        self.m2_dx = 0.0
        self.m2_dy = 0.0
        self.m2_Rx = 0.0
        self.m2_Ry = 0.0

        self.cam_dz = 0.0
        self.cam_dx = 0.0
        self.cam_dy = 0.0
        self.cam_Rx = 0.0
        self.cam_Ry = 0.0

        self.offsets = np.zeros(50)
        self.text = ""
        self._n_iter = 0

        # Controls
        kwargs = {'layout':{'width':'180px'}, 'style':{'description_width':'initial'}}
        self.m2_dz_control = ipywidgets.FloatText(value=self.m2_dz, description="M2 dz (µm)", step=10, **kwargs)
        self.m2_dx_control = ipywidgets.FloatText(value=self.m2_dx, description="M2 dx (µm)", step=500, **kwargs)
        self.m2_dy_control = ipywidgets.FloatText(value=self.m2_dy, description="M2 dy (µm)", step=500, **kwargs)
        self.m2_Rx_control = ipywidgets.FloatText(value=self.m2_Rx, description="M2 Rx (arcsec)", step=10, **kwargs)
        self.m2_Ry_control = ipywidgets.FloatText(value=self.m2_Ry, description="M2 Ry (arcsec)", step=10, **kwargs)
        self.cam_dz_control = ipywidgets.FloatText(value=self.cam_dz, description="Cam dz (µm)", step=10, **kwargs)
        self.cam_dx_control = ipywidgets.FloatText(value=self.cam_dx, description="Cam dx (µm)", step=2000, **kwargs)
        self.cam_dy_control = ipywidgets.FloatText(value=self.cam_dy, description="Cam dy (µm)", step=2000, **kwargs)
        self.cam_Rx_control = ipywidgets.FloatText(value=self.cam_Rx, description="Cam Rx (arcsec)", step=10, **kwargs)
        self.cam_Ry_control = ipywidgets.FloatText(value=self.cam_Ry, description="Cam Ry (arcsec)", step=10, **kwargs)
        self.zero_control = ipywidgets.Button(description="Zero")
        self.randomize_control = ipywidgets.Button(description="Randomize")
        self.reveal_control = ipywidgets.Button(description="Reveal")
        self.solve_control = ipywidgets.Button(description="Solve")
        self.control_truncated_control = ipywidgets.Button(description="Control w/ Trunc")
        self.control_penalty_control = ipywidgets.Button(description="Control w/ Penalty")

        self.controls = ipywidgets.VBox([
            self.m2_dz_control, self.m2_dx_control, self.m2_dy_control,
            self.m2_Rx_control, self.m2_Ry_control,
            self.cam_dz_control, self.cam_dx_control, self.cam_dy_control,
            self.cam_Rx_control, self.cam_Ry_control,
            self.zero_control, self.randomize_control,
            self.reveal_control, self.solve_control,
            self.control_truncated_control, self.control_penalty_control
        ])

        # Observers
        self.m2_dz_control.observe(lambda change: self.handle_event(change, 'm2_dz'), 'value')
        self.m2_dx_control.observe(lambda change: self.handle_event(change, 'm2_dx'), 'value')
        self.m2_dy_control.observe(lambda change: self.handle_event(change, 'm2_dy'), 'value')
        self.m2_Rx_control.observe(lambda change: self.handle_event(change, 'm2_Rx'), 'value')
        self.m2_Ry_control.observe(lambda change: self.handle_event(change, 'm2_Ry'), 'value')
        self.cam_dz_control.observe(lambda change: self.handle_event(change, 'cam_dz'), 'value')
        self.cam_dx_control.observe(lambda change: self.handle_event(change, 'cam_dx'), 'value')
        self.cam_dy_control.observe(lambda change: self.handle_event(change, 'cam_dy'), 'value')
        self.cam_Rx_control.observe(lambda change: self.handle_event(change, 'cam_Rx'), 'value')
        self.cam_Ry_control.observe(lambda change: self.handle_event(change, 'cam_Ry'), 'value')
        self.zero_control.on_click(self.zero)
        self.randomize_control.on_click(self.randomize)
        self.reveal_control.on_click(self.reveal)
        self.solve_control.on_click(self.solve)
        self.control_truncated_control.on_click(self.control_truncated)
        self.control_penalty_control.on_click(self.control_penalty)

        self.view = self._view()
        self.textout = ipywidgets.Textarea(
            value=self.text,
            layout=ipywidgets.Layout(height="250pt", width="auto")
        )
        self._pause_handler = False
        self._is_playing = False
        self._control_history = []

    def zero(self, b):
        self.m2_dz = 0.0
        self.m2_dx = 0.0
        self.m2_dy = 0.0
        self.m2_Rx = 0.0
        self.m2_Ry = 0.0
        self.cam_dz = 0.0
        self.cam_dx = 0.0
        self.cam_dy = 0.0
        self.cam_Rx = 0.0
        self.cam_Ry = 0.0
        self.offsets = np.zeros(50)
        self.text = 'Values Zeroed!'
        self._is_playing = False
        self._n_iter = 0
        self.update()
        self._control_history = []
        self._control_history.append(
            (
                self.wfe,
                self.m2_dz,
                self.m2_dx,
                self.m2_dy,
                self.m2_Rx,
                self.m2_Ry,
                self.cam_dz,
                self.cam_dx,
                self.cam_dy,
                self.cam_Rx,
                self.cam_Ry,
            )
        )

    def randomize(self, b):
        # amplitudes
        amp = [25.0, 1000.0, 1000.0, 25.0, 25.0, 25.0, 4000.0, 4000.0, 25.0, 25.0]
        offsets = self.offset_rng.normal(scale=amp)*2
        self.m2_dz = 0.0
        self.m2_dx = 0.0
        self.m2_dy = 0.0
        self.m2_Rx = 0.0
        self.m2_Ry = 0.0
        self.cam_dz = 0.0
        self.cam_dx = 0.0
        self.cam_dy = 0.0
        self.cam_Rx = 0.0
        self.cam_Ry = 0.0
        self.offsets = np.round(np.concatenate([offsets, np.zeros(40)]), 2)
        self.text = 'Values Randomized!'
        self._is_playing = True
        self._n_iter = 0
        self.update()
        self._control_history = []
        self._control_history.append(
            (
                self.wfe,
                self.m2_dz,
                self.m2_dx,
                self.m2_dy,
                self.m2_Rx,
                self.m2_Ry,
                self.cam_dz,
                self.cam_dx,
                self.cam_dy,
                self.cam_Rx,
                self.cam_Ry,
            )
        )

    def reveal(self, b):
        self.text = ""
        self.text += f"M2 dz: {self.offsets[0]:.2f} µm\n\n"
        self.text += f"M2 dx: {self.offsets[1]:.2f} µm\n\n"
        self.text += f"M2 dy: {self.offsets[2]:.2f} µm\n\n"
        self.text += f"M2 Rx: {self.offsets[3]:.2f} arcsec\n\n"
        self.text += f"M2 Ry: {self.offsets[4]:.2f} arcsec\n\n"
        self.text += f"Cam dz: {self.offsets[5]:.2f} µm\n\n"
        self.text += f"Cam dx: {self.offsets[6]:.2f} µm\n\n"
        self.text += f"Cam dy: {self.offsets[7]:.2f} µm\n\n"
        self.text += f"Cam Rx: {self.offsets[8]:.2f} arcsec\n\n"
        self.text += f"Cam Ry: {self.offsets[9]:.2f} arcsec\n\n"
        self._is_playing = False
        self.update()

    def solve(self, b):
        self._is_playing = False
        self.m2_dz = -self.offsets[0]
        self.m2_dx = -self.offsets[1]
        self.m2_dy = -self.offsets[2]
        self.m2_Rx = -self.offsets[3]
        self.m2_Ry = -self.offsets[4]
        self.cam_dz = -self.offsets[5]
        self.cam_dx = -self.offsets[6]
        self.cam_dy = -self.offsets[7]
        self.cam_Rx = -self.offsets[8]
        self.cam_Ry = -self.offsets[9]
        self.reveal(None)
        self.update()

    def control_truncated(self, b):
        # Don't use M2 tilt or camera piston.
        dz_fit = self.fit_dz()
        sens = np.array(self.sens)
        sens = sens[:, [0,1,2,6,7,8,9]]
        dof_fit, _, _, _ = lstsq(sens, dz_fit)
        dof_fit = np.round(dof_fit, 2)
        full_dof = np.zeros(10)
        full_dof[[0,1,2,6,7,8,9]] = dof_fit
        self.apply_dof(-full_dof)
        self._plot_control_history()

    def control_penalty(self, b):
        # Add rows to sens matrix to penalize large dof
        dz_fit = self.fit_dz()
        sens = np.zeros((9+10, 10))
        sens[:9, :] = self.sens
        alpha = 1e-12 # strength of penalty
        for i in range(9, 19):
            sens[i, i-9] = alpha
        dof_fit, _, _, _ = lstsq(sens, np.concatenate([dz_fit, [0]*10]))
        dof_fit = np.round(dof_fit, 2)
        self.apply_dof(-dof_fit)
        self._plot_control_history()

    def _plot_control_history(self):
        self._control_history.append(
            (
                self.wfe,
                self.m2_dz,
                self.m2_dx,
                self.m2_dy,
                self.m2_Rx,
                self.m2_Ry,
                self.cam_dz,
                self.cam_dx,
                self.cam_dy,
                self.cam_Rx,
                self.cam_Ry,
            )
        )
        with self.control_log:
            with contextlib.suppress(AttributeError):
                fig, axes = plt.subplots(nrows=1, ncols=4, figsize=(8, 2))
                axes[0].plot([x[0] for x in self._control_history], c='k')
                axes[1].plot([x[1] for x in self._control_history], c='b')
                axes[1].plot([x[6] for x in self._control_history], c='r')
                axes[2].plot([x[2] for x in self._control_history], c='b', ls='--')
                axes[2].plot([x[3] for x in self._control_history], c='b', ls=':')
                axes[2].plot([x[7] for x in self._control_history], c='r', ls='--')
                axes[2].plot([x[8] for x in self._control_history], c='r', ls=':')
                axes[3].plot([x[4] for x in self._control_history], c='b', ls='--')
                axes[3].plot([x[5] for x in self._control_history], c='b', ls=':')
                axes[3].plot([x[9] for x in self._control_history], c='r', ls='--')
                axes[3].plot([x[10] for x in self._control_history], c='r', ls=':')

                axes[0].set_ylabel("WFE (µm)")
                axes[1].set_ylabel("dz (µm)")
                axes[2].set_ylabel("dx, dy (µm)")
                axes[3].set_ylabel("Rx, Ry (arcsec)")
                for ax in axes:
                    ax.set_xlabel("Iteration")

                fig.tight_layout()
                plt.show(fig)

    def fit_dz(self):
        # Wavefront estimation part of the control loop.
        Rubin_obsc = yaml.safe_load(open(Path(danish.datadir)/'RubinObsc.yaml'))
        factory = danish.DonutFactory(
            R_outer=4.18, R_inner=2.5498,
            obsc_radii=Rubin_obsc['radii'], obsc_motion=Rubin_obsc['motion'],
            focal_length=10.31, pixel_scale=10e-6
        )
        sky_level = 1.0

        dz_terms = (
            (1, 4),                          # defocus
            (2, 4), (3, 4),                  # field tilt
            (2, 5), (3, 5), (2, 6), (3, 6),  # linear astigmatism
            (1, 7), (1, 8),                  # constant coma
            # (1, 9), (1, 10),                 # constant trefoil
            # (1, 11),                         # constant spherical
            # (1, 12), (1, 13),                # second astigmatism
            # (1, 14), (1, 15),                # quatrefoil
            # (1, 16), (1, 17),
            # (1, 18), (1, 19),
            # (1, 20), (1, 21),
            # (1, 22)
        )

        thxs = []
        thys = []
        z_refs = []
        imgs = []
        names = []
        for raft in self._rafts.values():
            if raft.name.startswith('R'):
                continue
            thxs.append(np.deg2rad(raft.thx))
            thys.append(np.deg2rad(raft.thy))
            z_refs.append(raft.z_ref)
            imgs.append(raft.img.get_array().data[::-1, ::-1])
            names.append(raft.name)

        fitter = danish.MultiDonutModel(
            factory, z_refs=z_refs, dz_terms=dz_terms,
            field_radius=np.deg2rad(1.8), thxs=thxs, thys=thys
        )
        nstar = len(thxs)
        guess = [0.0]*nstar + [0.0]*nstar + [0.5] + [0.0]*len(dz_terms)
        sky_levels = [sky_level]*nstar

        with self.control_log:
            print()
            result = least_squares(
                fitter.chi, guess, jac=fitter.jac,
                ftol=1e-3, xtol=1e-3, gtol=1e-3,
                max_nfev=20, verbose=2,
                args=(imgs, sky_levels)
            )

        dxs_fit, dys_fit, fwhm_fit, dz_fit = fitter.unpack_params(result.x)

        with self.control_log:
            with contextlib.suppress(AttributeError):
                mods = fitter.model(
                    dxs_fit, dys_fit, fwhm_fit, dz_fit
                )
                fig, axes = plt.subplots(nrows=2, ncols=8, figsize=(8, 2))
                for i in range(8):
                    axes[0,i].imshow(imgs[i]/np.sum(imgs[i]))
                    axes[1,i].imshow(mods[i]/np.sum(mods[i]))
                    axes[0,i].set_title(names[i])
                for ax in axes.ravel():
                    ax.set_xticks([])
                    ax.set_yticks([])
                    ax.set_aspect('equal')
                axes[0, 0].set_ylabel('Data')
                axes[1, 0].set_ylabel('Model')
                fig.tight_layout()
                plt.show(fig)

            print()
            print("Found double Zernikes:")
            for z_term, val in zip(dz_terms, dz_fit):
                print(f"DZ({z_term[0]}, {z_term[1]}): {val*1e9:10.2f} nm")

        return dz_fit

    @cached_property
    def sens(self):
        dz_terms = (
            (1, 4),                          # defocus
            (2, 4), (3, 4),                  # field tilt
            (2, 5), (3, 5), (2, 6), (3, 6),  # linear astigmatism
            (1, 7), (1, 8),                  # constant coma
        )
        sens = np.zeros((9, 10))
        dz_ref = batoid.doubleZernike(
            self.fiducial, np.deg2rad(1.8), self.wavelength,
            jmax=9, kmax=4, eps=0.61
        )*self.wavelength
        for idof in range(10):
            dof = np.zeros(50)
            dof[idof] = 100.0  # microns or arcsec
            builder = self.builder.with_aos_dof(dof.tolist())
            telescope = builder.build()
            dz_p = batoid.doubleZernike(
                telescope, np.deg2rad(1.8), self.wavelength,
                jmax=9, kmax=4, eps=0.61
            )*self.wavelength
            for i, (j, k) in enumerate(dz_terms):
                sens[i, idof] = (dz_p - dz_ref)[j, k]/100
        return sens

    def apply_dof(self, dof):
        with self.control_log:
            print()
            print("Applying DOF:")
            print(f"M2 dz:  {dof[0]:10.2f} µm")
            print(f"M2 dx:  {dof[1]:10.2f} µm")
            print(f"M2 dy:  {dof[2]:10.2f} µm")
            print(f"M2 Rx:  {dof[3]:10.2f} arcsec")
            print(f"M2 Ry:  {dof[4]:10.2f} arcsec")
            print(f"cam dz: {dof[5]:10.2f} µm")
            print(f"cam dx: {dof[6]:10.2f} µm")
            print(f"cam dy: {dof[7]:10.2f} µm")
            print(f"cam Rx: {dof[8]:10.2f} arcsec")
            print(f"cam Ry: {dof[9]:10.2f} arcsec")
        self._is_playing = False
        self.m2_dz += dof[0]
        self.m2_dx += dof[1]
        self.m2_dy += dof[2]
        self.m2_Rx += dof[3]
        self.m2_Ry += dof[4]
        self.cam_dz += dof[5]
        self.cam_dx += dof[6]
        self.cam_dy += dof[7]
        self.cam_Rx += dof[8]
        self.cam_Ry += dof[9]

        self.update()

    def handle_event(self, change, attr):
        if self._pause_handler:
            return
        setattr(self, attr, change['new'])
        if self._is_playing:
            self._n_iter += 1
        self.update()

    def _view(self):
        self._fig = fig = plt.figure(constrained_layout=True, figsize=(5, 5))
        raftspec = [[  None, "in04",  None,  None,  None,   None,   None],
                    [  None, "ex04", "R14", "R24", "R34", "ex44", "in44"],
                    [  None,  "R03", "R13", "R23", "R33",  "R43",   None],
                    [  None,  "R02", "R12", "R22", "R32",  "R42",   None],
                    [  None,  "R01", "R11", "R21", "R31",  "R41",   None],
                    ["in00", "ex00", "R00", "R10", "R20", "ex40",   None],
                    [  None,   None,  None,  None,  None, "in40",   None]]
        self._axes = fig.subplot_mosaic(
            raftspec, empty_sentinel=None
        )

        # Determine spacing
        center = (self._axes["R22"].transAxes + fig.transFigure.inverted()).transform([0.5, 0.5])
        r02 = (self._axes["R02"].transAxes + fig.transFigure.inverted()).transform([0.5, 0.5])
        dx = r02[0] - center[0]  # make this 1.4 degrees
        factor = 1.4/dx

        self._rafts = {}
        for k, ax in self._axes.items():
            ax.set_xticks([])
            ax.set_yticks([])

            mytrans = ax.transAxes + fig.transFigure.inverted()
            x, y = mytrans.transform([0.5, 0.5])
            if "R" in k:
                nphot = 1000
                nx = 21
                thx=(x-center[0])*factor
                thy=(y-center[1])*factor
            else:
                nphot = 50000
                nx = 181
            # Redo WF centering
            if k == "in00":
                thx, thy = -5.25*3.5/15, -5*3.5/15
            elif k == "ex00":
                thx, thy = -4.75*3.5/15, -5*3.5/15

            elif k == "in04":
                thx, thy = -5*3.5/15, 5.25*3.5/15
            elif k == "ex04":
                thx, thy = -5*3.5/15, 4.75*3.5/15

            elif k == "in40":
                thx, thy = 5*3.5/15, -5.25*3.5/15
            elif k == "ex40":
                thx, thy = 5*3.5/15, -4.75*3.5/15

            elif k == "in44":
                thx, thy = 5.25*3.5/15, 5*3.5/15
            elif k == "ex44":
                thx, thy = 4.75*3.5/15, 5*3.5/15

            self._rafts[k] = Raft(
                k, thx, thy, nphot=nphot, fiducial=self.fiducial,
                img=ax.imshow(np.zeros((nx, nx)), vmin=0, vmax=1)
            )

            ax.text(0.02, 0.87, k, transform=ax.transAxes, fontsize=6, color='white')

        # self._axes["in04"].text(0.05, 0.9, "InR04", transform=self._axes["in04"].transAxes, fontsize=6, color='white')
        # self._axes["ex04"].text(0.05, 0.9, "ExR04", transform=self._axes["ex04"].transAxes, fontsize=6, color='white')

        self.wfe_text = fig.text(0.31, 0.89, "WFE", ha="left", va="center", fontsize=16)
        self.win_text = fig.text(0.31, 0.96, "", ha="left", va="center", fontsize=16, color='red')

        self._canvas = fig.canvas
        self._canvas.header_visible = False

        out = ipywidgets.Output()
        with out:
            plt.show(fig)
        return out

    def update(self):
        dof = [self.m2_dz, self.m2_dx, self.m2_dy, self.m2_Rx, self.m2_Ry]
        dof += [self.cam_dz, self.cam_dx, self.cam_dy, self.cam_Rx, self.cam_Ry]
        dof += [0]*40
        dof = (self.offsets + dof).tolist()

        builder = self.builder.with_aos_dof(dof)
        telescope = builder.build()

        self.dz = batoid.doubleZernike(
            telescope,
            np.deg2rad(1.75),
            500e-9,
            kmax=15,
            jmax=37,
            eps=0.61
        )
        self.wfe = np.sqrt(np.sum(np.square(self.dz[:, 4:])))
        self.wfe *= 500e-9*1e6  # microns

        for raft in self._rafts.values():
            raft.draw(telescope, seeing=0.5, rng=self.rng)
        self.wfe_text.set_text(f"WFE = {self.wfe:.3f} µm   iter: {self._n_iter}")
        if self._is_playing:
            if self.wfe < 0.5:
                self.win_text.set_text("You Win!")
                self._is_playing = False
        self._canvas.draw()

        self.textout.value = self.text

        self._pause_handler = True
        self.m2_dz_control.value = self.m2_dz
        self.m2_dx_control.value = self.m2_dx
        self.m2_dy_control.value = self.m2_dy
        self.m2_Rx_control.value = self.m2_Rx
        self.m2_Ry_control.value = self.m2_Ry
        self.cam_dz_control.value = self.cam_dz
        self.cam_dx_control.value = self.cam_dx
        self.cam_dy_control.value = self.cam_dy
        self.cam_Rx_control.value = self.cam_Rx
        self.cam_Ry_control.value = self.cam_Ry
        self._pause_handler = False

    def display(self):
        from IPython.display import display
        self.app = ipywidgets.HBox([
            self.view,
            self.controls,
            self.textout

        ])

        display(self.app)
        self.update()



from pathlib import Path
from collections import namedtuple
from copy import copy
from functools import lru_cache
from numbers import Real

import astropy.io.fits as fits
import batoid
import galsim
import numpy as np
import yaml

from .utils import _node_to_grid, _fits_cache, attach_attr


BendingMode = namedtuple(
    "BendingMode",
    "zk R_outer R_inner x y z dzdx dzdy d2zdxy"
)


RealizedBend = namedtuple(
    "RealizedBend",
    "zk grid"
)


@lru_cache(maxsize=2)
def m1m3_fea_nodes(fea_dir):
    data = fits.getdata(Path(fea_dir) / "M1M3_1um_156_grid.fits.gz")
    idx = data[:, 0]
    bx = data[:, 1]  # (5256,)
    by = data[:, 2]
    idx1 = (idx == 1)
    idx3 = (idx == 3)
    bx.flags.writeable = False
    by.flags.writeable = False
    idx1.flags.writeable = False
    idx3.flags.writeable = False
    return bx, by, idx1, idx3


@lru_cache(maxsize=4)
def m1m3_grid_xy(bend_dir):
    with open(Path(bend_dir) / "bend.yaml") as f:
        config = yaml.safe_load(f)
    m1_grid_xy = fits.getdata(Path(bend_dir) / config['M1']['grid']['coords'])
    m3_grid_xy = fits.getdata(Path(bend_dir) / config['M3']['grid']['coords'])
    m1_grid_xy.flags.writeable = False
    m3_grid_xy.flags.writeable = False
    return m1_grid_xy, m3_grid_xy


@lru_cache(maxsize=2)
def m2_fea_nodes(fea_dir):
    data = fits.getdata(Path(fea_dir) / "M2_1um_grid.fits.gz")
    bx = data[:, 1]  # meters
    by = data[:, 2]
    bx.flags.writeable = False
    by.flags.writeable = False
    return bx, by


@lru_cache(maxsize=4)
def m2_grid_xy(bend_dir):
    with open(Path(bend_dir) / "bend.yaml") as f:
        config = yaml.safe_load(f)
    m2_grid_xy = fits.getdata(Path(bend_dir) / config['M2']['grid']['coords'])
    m2_grid_xy.flags.writeable = False
    return m2_grid_xy


@lru_cache(maxsize=16)
def m1m3_gravity(fea_dir, optic, zenith):
    zdata = _fits_cache(fea_dir, "M1M3_dxdydz_zenith.fits.gz")
    hdata = _fits_cache(fea_dir, "M1M3_dxdydz_horizon.fits.gz")

    if zenith is None:
        return np.zeros_like(zdata[:,2])

    dxyz = (
        zdata * np.cos(zenith) +
        hdata * np.sin(zenith)
    )
    dz = dxyz[:,2]

    # Interpolate these node displacements into z-displacements at
    # original node x/y positions.
    bx, by, idx1, idx3 = m1m3_fea_nodes(fea_dir)

    # M1
    zRef = optic['M1'].surface.sag(bx[idx1], by[idx1])
    zpRef = optic['M1'].surface.sag(
        (bx+dxyz[:, 0])[idx1],
        (by+dxyz[:, 1])[idx1]
    )
    dz[idx1] += zRef - zpRef

    # M3
    zRef = optic['M3'].surface.sag(bx[idx3], by[idx3])
    zpRef = optic['M3'].surface.sag(
        (bx+dxyz[:, 0])[idx3],
        (by+dxyz[:, 1])[idx3]
    )
    dz[idx3] += zRef - zpRef

    # Subtract PTT
    # This kinda makes sense for M1, but why for combined M1M3?
    zBasis = galsim.zernike.zernikeBasis(
        3, bx, by, R_outer=4.18, R_inner=2.558
    )
    coefs, *_ = np.linalg.lstsq(zBasis.T, dxyz[:, 2], rcond=None)
    zern = galsim.zernike.Zernike(coefs, R_outer=4.18, R_inner=2.558)
    dz -= zern(bx, by)

    dz.flags.writeable = False
    return dz


@lru_cache(maxsize=16)
def m1m3_temperature(fea_dir, TBulk, TxGrad, TyGrad, TzGrad, TrGrad):
    tbdz, txdz, tydz, tzdz, trdz = _fits_cache(
        fea_dir,
        "M1M3_thermal_FEA.fits.gz"
    )

    out = TBulk * tbdz
    out += TxGrad * txdz
    out += TyGrad * tydz
    out += TzGrad * tzdz
    out += TrGrad * trdz
    out *= 1e-6
    out.flags.writeable = False
    return out


@lru_cache(maxsize=16)
def m1m3_lut(fea_dir, zenith, error, seed):
    if zenith is None:
        return np.zeros(256)

    from scipy.interpolate import interp1d
    data = _fits_cache(fea_dir, "M1M3_LUT.fits.gz")

    LUT_force = interp1d(data[0], data[1:])(zenith.deg)

    if error != 0.0:
        # Get current forces so we can rebalance after applying random error
        z_force = np.sum(LUT_force[:156])
        y_force = np.sum(LUT_force[156:])

        rng = np.random.default_rng(seed)
        LUT_force *= rng.uniform(1-error, 1+error, size=len(LUT_force))

        # Balance forces by adjusting means in 2 ranges
        LUT_force[:156] -= np.mean(LUT_force[:156]) - z_force
        LUT_force[156:] -= np.mean(LUT_force[156:]) - y_force

        # # Balance forces by manipulating these 2 actuators
        # LUT_force[155] = z_force - np.sum(LUT_force[:155])
        # LUT_force[-1] = y_force - np.sum(LUT_force[156:-1])

    zf = _fits_cache(fea_dir, "M1M3_force_zenith.fits.gz")
    hf = _fits_cache(fea_dir, "M1M3_force_horizon.fits.gz")
    u0 = zf * np.cos(zenith)
    u0 += hf * np.sin(zenith)

    out = LUT_force - u0
    out.flags.writeable = False
    return out


def m1m3_force_to_surface(fea_dir, forces):
    G = _fits_cache(fea_dir, "M1M3_influence_256.fits.gz")
    out = np.dot(G, forces)
    out.flags.writeable = False
    return out


def transform_zernike(zernike, R_outer, R_inner):
    xy = zernike._coef_array_xy
    ret = galsim.zernike.Zernike.__new__(galsim.zernike.Zernike)
    ret._coef_array_xy = xy
    ret.R_outer = R_outer
    ret.R_inner = R_inner
    return ret


def _load_mirror_bend(bend_dir, inds, config):
    zk = fits.getdata(Path(bend_dir) / config['zk']['file'])
    grid = fits.getdata(Path(bend_dir) / config['grid']['file'])
    coords = fits.getdata(Path(bend_dir) / config['grid']['coords'])
    inds = np.array(inds)
    zk = zk[inds]
    grid = grid[:, inds]
    zk.flags.writeable = False
    grid.flags.writeable = False
    coords.flags.writeable = False
    return BendingMode(
        zk, config['zk']['R_outer'], config['zk'].get('R_inner', 0.0),
        coords[0], coords[1],
        grid[0], grid[1], grid[2], grid[3]
    )


@lru_cache(maxsize=4)
def load_bend(bend_dir, inds):
    with open(Path(bend_dir) / "bend.yaml") as f:
        config = yaml.safe_load(f)
    m1 = _load_mirror_bend(bend_dir, inds, config['M1'])
    m2 = _load_mirror_bend(bend_dir, inds, config['M2'])
    m3 = _load_mirror_bend(bend_dir, inds, config['M3'])
    return dict(M1=m1, M2=m2, M3=m3)


@lru_cache(maxsize=16*3)
def realize_bend(bend_dir, dof, inds, mirror):
    modes = load_bend(bend_dir, inds)[mirror]
    dof = np.array(dof)
    zk = galsim.zernike.Zernike(
        np.tensordot(dof, modes.zk, axes=1),
        R_outer=modes.R_outer,
        R_inner=modes.R_inner
    )
    z = np.tensordot(dof, modes.z, axes=1)
    dzdx = np.tensordot(dof, modes.dzdx, axes=1)
    dzdy = np.tensordot(dof, modes.dzdy, axes=1)
    d2zdxy = np.tensordot(dof, modes.d2zdxy, axes=1)
    grid = np.stack([z, dzdx, dzdy, d2zdxy])
    grid.flags.writeable = False
    return RealizedBend(zk, grid)


@lru_cache(maxsize=16)
def m2_gravity(fea_dir, zenith):
    data = _fits_cache(fea_dir, "M2_GT_FEA.fits.gz")
    if zenith is None:
        return np.zeros_like(data[0])

    zdz, hdz = data[0:2]

    out = zdz * (np.cos(zenith) - 1)
    out += hdz * np.sin(zenith)
    out *= 1e-6  # micron -> meters
    out.flags.writeable = False
    return out


@lru_cache(maxsize=16)
def m2_temperature(fea_dir, TzGrad, TrGrad):
    data = _fits_cache(fea_dir, "M2_GT_FEA.fits.gz")
    tzdz, trdz = data[2:4]

    out = TzGrad * tzdz
    out += TrGrad * trdz
    out *= 1e-6
    out.flags.writeable = False
    return out


@lru_cache(maxsize=16)
def LSSTCam_gravity(fea_dir, zenith, rotation):
    if zenith is None:
        return None

    camera_gravity_zk = {}
    cam_data = [
        ('L1S1', 'L1_entrance'),
        ('L1S2', 'L1_exit'),
        ('L2S1', 'L2_entrance'),
        ('L2S2', 'L2_exit'),
        ('L3S1', 'L3_entrance'),
        ('L3S2', 'L3_exit')
    ]
    for tname, bname in cam_data:
        data = _fits_cache(fea_dir, tname+"zer.fits.gz")
        grav_zk = data[0, 3:] * (np.cos(zenith) - 1)
        grav_zk += (
            data[1, 3:] * np.cos(rotation) +
            data[2, 3:] * np.sin(rotation)
        ) * np.sin(zenith)

        # remap Andy -> Noll Zernike indices
        zIdxMapping = [
            1, 3, 2, 5, 4, 6, 8, 9, 7, 10, 13, 14, 12, 15, 11, 19, 18, 20,
            17, 21, 16, 25, 24, 26, 23, 27, 22, 28
        ]
        grav_zk = grav_zk[[x - 1 for x in zIdxMapping]]
        grav_zk *= 1e-3  # mm -> m
        # tsph -> batoid 0-index offset
        grav_zk = np.concatenate([[0], grav_zk])
        camera_gravity_zk[bname] = grav_zk
    return camera_gravity_zk


@lru_cache(maxsize=16)
def LSSTCam_temperature(fea_dir, TBulk):
    camera_temperature_zk = {}
    cam_data = [
        ('L1S1', 'L1_entrance'),
        ('L1S2', 'L1_exit'),
        ('L2S1', 'L2_entrance'),
        ('L2S2', 'L2_exit'),
        ('L3S1', 'L3_entrance'),
        ('L3S2', 'L3_exit')
    ]
    for tname, bname in cam_data:
        data = _fits_cache(fea_dir, tname+"zer.fits.gz")
        # subtract pre-compensated grav...
        TBulk = np.clip(
            TBulk,
            np.min(data[3:, 2])+0.001,
            np.max(data[3:, 2])-0.001
        )
        fidx = np.interp(TBulk, data[3:, 2], np.arange(len(data[3:, 2])))+3
        idx = int(np.floor(fidx))
        whi = fidx - idx
        wlo = 1 - whi
        temp_zk = wlo * data[idx, 3:] + whi * data[idx+1, 3:]

        # subtract reference temperature zk (0 deg C is idx=5)
        temp_zk -= data[5, 3:]

        # remap Andy -> Noll Zernike indices
        zIdxMapping = [
            1, 3, 2, 5, 4, 6, 8, 9, 7, 10, 13, 14, 12, 15, 11, 19, 18, 20,
            17, 21, 16, 25, 24, 26, 23, 27, 22, 28
        ]
        temp_zk = temp_zk[[x - 1 for x in zIdxMapping]]
        temp_zk *= 1e-3  # mm -> m
        # tsph -> batoid 0-index offset
        temp_zk = np.concatenate([[0], temp_zk])
        camera_temperature_zk[bname] = temp_zk


class LSSTBuilder:
    def __init__(
        self,
        fiducial,
        fea_dir="fea_legacy",
        bend_dir="bend",
        use_m1m3_modes=None,
        use_m2_modes=None,
    ):
        """Create a Simony Survey Telescope with LSSTCam camera builder.

        Parameters
        ----------
        fiducial : batoid.Optic
            Optic before finite-element analysis (FEA) or active optics system
            (AOS) perturbations are applied.
        fea_dir : str
            Directory containing the FEA files.
        bend_dir : str
            Directory containing the bending mode files.
        use_m1m3_modes : list of int, optional
            List indicating the ordering of the bending modes to use for M1M3.
            For example, [9, 0, 4] would mean use the 9th, 0th, and 4th modes
            from the bending mode directory, but map these to the 10th, 11th,
            and 12th indices of the AOS degrees-of-freedom (The 0th through 9th
            indices are always reserved for the rigid body degrees of freedom of
            M2 and the camera).  If None, then use the modes in the order
            specified in the bending mode directory.
        use_m2_modes : list of int, optional
            List indicating the ordering of the bending modes to use for M2.
            Note that the M2 bending mode indices always come immediately after
            the M1M3 bending mode indices.  So if use_m1m3_modes is [9, 0, 4]
            and use_m2_modes is [4, 2], then the 10th, 11th, and 12th AOS
            degrees of freedom will be the 9th, 0th, and 4th M1M3 bending modes,
            and the 13th and 14th AOS degrees of freedom will be the 4th and 2nd
            M2 bending modes.  If None, then use the modes in the order
            specified in the bending mode directory.
        """
        # Number of FEA nodes and actuators is inferred from content of fea_dir.
        # Number of bending modes is inferred from content of bend_dir.
        # These can effect the sizes of intermediate numpy arrays, and also the
        # required inputs for the with_aos_dof and with_*_forces methods.

        self.fiducial = fiducial
        fea_dir = Path(fea_dir)
        bend_dir = Path(bend_dir)

        if not fea_dir.is_dir():
            # See if we can find the data in the batoid_rubin data directory.
            from . import datadir
            fea_dir = datadir / fea_dir
            if not fea_dir.is_dir():
                # See if we can download the data from Zenodo.
                from .data.download_rubin_data import (
                    download_rubin_data, zenodo_dois
                )
                if fea_dir.name in zenodo_dois:
                    args = namedtuple("Args", ["dataset", "outdir"])
                    args.dataset = fea_dir.parts[-1]
                    args.outdir = None
                    download_rubin_data(args)
                else:
                    raise ValueError("Cannot infer fea_dir.")

        if not bend_dir.is_dir():
            # See if we can find the data in the batoid_rubin data directory.
            from . import datadir
            bend_dir = datadir / bend_dir
            if not bend_dir.is_dir():
                # See if we can download the data from Zenodo.
                from .data.download_rubin_data import (
                    download_rubin_data, zenodo_dois
                )
                if bend_dir.name in zenodo_dois:
                    args = namedtuple("Args", ["dataset", "outdir"])
                    args.dataset = bend_dir.parts[-1]
                    args.outdir = None
                    download_rubin_data(args)
                else:
                    raise ValueError("Cannot infer bend_dir.")

        self.fea_dir = fea_dir
        self.bend_dir = bend_dir

        if use_m1m3_modes is None:
            with open(bend_dir / "bend.yaml") as f:
                config = yaml.safe_load(f)
                use_m1m3_modes = config['use_m1m3_modes']
        self.use_m1m3_modes = use_m1m3_modes
        self.m1m3_dof_indices = slice(10, 10+len(self.use_m1m3_modes))

        if use_m2_modes is None:
            with open(bend_dir / "bend.yaml") as f:
                config = yaml.safe_load(f)
                use_m2_modes = config['use_m2_modes']
        self.use_m2_modes = use_m2_modes
        self.m2_dof_indices = slice(
            self.m1m3_dof_indices.stop,
            self.m1m3_dof_indices.stop+len(self.use_m2_modes),
        )

        if 'LSST.LSSTCamera' in self.fiducial.itemDict:
            self.cam_name = 'LSSTCamera'
        elif 'ComCam.ComCam' in self.fiducial.itemDict:
            self.cam_name = 'ComCam.ComCam'
        else:
            raise ValueError("Unsupported optic")

        # "Input" variables.
        self.m1m3_zenith = None
        self.m1m3_TBulk = 0.0
        self.m1m3_TxGrad = 0.0
        self.m1m3_TyGrad = 0.0
        self.m1m3_TzGrad = 0.0
        self.m1m3_TrGrad = 0.0
        self.m1m3_lut_zenith = None
        self.m1m3_lut_error = 0.0
        self.m1m3_lut_seed = 1
        self.m1m3_extra_forces = np.zeros(256)

        self.m2_zenith = None
        self.m2_TzGrad = 0.0
        self.m2_TrGrad = 0.0

        self.camera_zenith = None
        self.camera_rotation = None
        self.camera_TBulk = None

        self.dof = np.zeros(50)
        self.extra_zk = None
        self.extra_zk_eps = None

    @attach_attr(
        _req_params={"zenith":galsim.Angle}
    )
    def with_m1m3_gravity(self, zenith):
        """Return new SSTBuilder that includes gravitational flexure of M1M3.

        Parameters
        ----------
        zenith : float
            Zenith angle in radians

        Returns
        -------
        ret : SSTBuilder
            New builder with M1M3 gravitation flexure applied.
        """
        if isinstance(zenith, Real):
            zenith = zenith * galsim.radians

        ret = copy(self)
        ret.m1m3_zenith = zenith
        return ret

    @attach_attr(
        _req_params={
            "m1m3_TBulk":float,
        },
        _opt_params={
            "m1m3_TxGrad":float,
            "m1m3_TyGrad":float,
            "m1m3_TzGrad":float,
            "m1m3_TrGrad":float,
        }
    )
    def with_m1m3_temperature(
        self,
        m1m3_TBulk,
        m1m3_TxGrad=0.0,
        m1m3_TyGrad=0.0,
        m1m3_TzGrad=0.0,
        m1m3_TrGrad=0.0,
    ):
        """Return new SSTBuilder that includes temperature flexure of M1M3.

        Parameters
        ----------
        m1m3_TBulk : float
            Bulk temperature in C.
        m1m3_TxGrad : float, optional
            Temperature gradient in x direction in C / m (?)
        m1m3_TyGrad : float, optional
            Temperature gradient in y direction in C / m (?)
        m1m3_TzGrad : float, optional
            Temperature gradient in z direction in C / m (?)
        m1m3_TrGrad : float, optional
            Temperature gradient in r direction in C / m (?)

        Returns
        -------
        ret : SSTBuilder
            New builder with M1M3 temperature flexure applied.
        """
        ret = copy(self)
        ret.m1m3_TBulk = m1m3_TBulk
        ret.m1m3_TxGrad = m1m3_TxGrad
        ret.m1m3_TyGrad = m1m3_TyGrad
        ret.m1m3_TzGrad = m1m3_TzGrad
        ret.m1m3_TrGrad = m1m3_TrGrad
        return ret

    @attach_attr(
        _req_params={
            "zenith":galsim.Angle,
        },
        _opt_params={
            "error":float,
            "seed":int,
        }
    )
    def with_m1m3_lut(self, zenith, error=0.0, seed=1):
        """Return new SSTBuilder that includes LUT perturbations of M1M3.

        Parameters
        ----------
        zenith : float
            Zenith angle in radians
        error : float, optional
            Fractional error to apply to LUT forces.

        Returns
        -------
        ret : SSTBuilder
            New builder with M1M3 LUT applied.
        """
        if isinstance(zenith, Real):
            zenith = zenith * galsim.radians

        ret = copy(self)
        ret.m1m3_lut_zenith = zenith
        ret.m1m3_lut_error=error
        ret.m1m3_lut_seed=seed
        return ret

    @attach_attr(
        _req_params={"zenith":galsim.Angle},
    )
    def with_m2_gravity(self, zenith):
        """Return new SSTBuilder that includes gravitational flexure of M2.

        Parameters
        ----------
        zenith : float
            Zenith angle in radians

        Returns
        -------
        ret : SSTBuilder
            New builder with M2 gravitation flexure applied.
        """
        if isinstance(zenith, Real):
            zenith = zenith * galsim.radians

        ret = copy(self)
        ret.m2_zenith = zenith
        return ret

    @attach_attr(
        _opt_params={
            "m2_TzGrad":float,
            "m2_TrGrad":float,
        }
    )
    def with_m2_temperature(
        self,
        m2_TzGrad=0.0,
        m2_TrGrad=0.0,
    ):
        """Return new SSTBuilder that includes temperature flexure of M2.

        Parameters
        ----------
        m2_TzGrad : float, optional
            Temperature gradient in z direction in C / m (?)
        m2_TrGrad : float, optional
            Temperature gradient in r direction in C / m (?)

        Returns
        -------
        ret : SSTBuilder
            New builder with M2 temperature flexure applied.
        """
        ret = copy(self)
        ret.m2_TzGrad = m2_TzGrad
        ret.m2_TrGrad = m2_TrGrad
        return ret

    @attach_attr(
        _req_params={
            "zenith":galsim.Angle,
            "rotation":galsim.Angle,
        },
    )
    def with_camera_gravity(self, zenith, rotation):
        """Return new SSTBuilder that includes gravitational flexure of camera.

        Parameters
        ----------
        zenith : float
            Zenith angle in radians
        rotation : float
            Rotation angle in radians

        Returns
        -------
        ret : SSTBuilder
            New builder with camera gravitation flexure applied.
        """
        if isinstance(zenith, Real):
            zenith = zenith * galsim.radians
        if isinstance(rotation, Real):
            rotation = rotation * galsim.radians

        ret = copy(self)
        ret.camera_zenith = zenith
        ret.camera_rotation = rotation
        return ret

    @attach_attr(
        _req_params={
            "camera_TBulk":float,
        },
    )
    def with_camera_temperature(self, camera_TBulk):
        """Return new SSTBuilder that includes temperature flexure of camera.

        Parameters
        ----------
        camera_TBulk : float
            Camera temperature in C

        Returns
        -------
        ret : SSTBuilder
            New builder with camera temperature flexure applied.
        """
        ret = copy(self)
        ret.camera_TBulk = camera_TBulk
        return ret

    @attach_attr(
        _req_params={
            "dof":None,
        },
    )
    def with_aos_dof(self, dof):
        """Return new SSTBuilder that includes specified AOS degrees of freedom

        Parameters
        ----------
        dof : ndarray (50,)
            AOS degrees of freedom.
            0,1,2 are M2 z,x,y in micron
            3,4 are M2 rot around x, y in arcsec
            5,6,7 are camera z,x,y in micron
            8,9 are camera rot around x, y in arcsec
            The next len(use_m1m3_modes) are M1M3 bending modes in micron
            The next len(use_m2_modes) are M2 bending modes in micron

        Returns
        -------
        ret : SSTBuilder
            New builder with specified AOS DOF.
        """
        assert len(dof) == 10+len(self.use_m1m3_modes)+len(self.use_m2_modes)
        ret = copy(self)
        ret.dof = dof
        return ret

    @attach_attr(
        _opt_params={
            "dof":None,
            "dx":float,
            "dy":float,
            "dz":float,
            "rx":galsim.Angle,
            "ry":galsim.Angle,
        },
    )
    def with_m2_rigid(
        self, dof=None, dx=None, dy=None, dz=None, rx=None, ry=None
    ):
        """Return new SSTBuilder that includes specified M2 rigid body DOF

        Note that you cannot specify both dof and dx,dy,dz,rx,ry.
        Values specified here overwrite M2 rigid body values specified in
        with_aos_dof, but not other degrees of freedom.

        Parameters
        ----------
        dof : ndarray (5,), optional
            0,1,2 are M2 z,x,y in micron
            3,4 are M2 rot around x, y in arcsec
        dx, dy, dz : float
            M2 displacement in micron
        rx, ry : galsim.Angle
            M2 rotation around x, y

        Returns
        -------
        ret : SSTBuilder
            New builder with specified M2 rigid body DOF.
        """
        ret = copy(self)
        if (
            any([x is not None for x in [dx, dy, dz, rx, ry]])
            and dof is not None
        ):
            raise ValueError("Cannot specify both dof and dx,dy,dz,rx,ry")
        if dof is not None:
            ret.dof[:5] = dof
        else:
            if dz is not None:
                ret.dof[0] = dz
            if dx is not None:
                ret.dof[1] = dx
            if dy is not None:
                ret.dof[2] = dy
            if rx is not None:
                ret.dof[3] = rx.deg*3600
            if ry is not None:
                ret.dof[4] = ry.deg*3600
        return ret


    @attach_attr(
        _opt_params={
            "dof":None,
            "dx":float,
            "dy":float,
            "dz":float,
            "rx":galsim.Angle,
            "ry":galsim.Angle,
        },
    )
    def with_camera_rigid(
        self, dof=None, dx=None, dy=None, dz=None, rx=None, ry=None
    ):
        """Return new SSTBuilder that includes specified camera rigid body DOF

        Note that you cannot specify both dof and dx,dy,dz,rx,ry.
        Values specified here overwrite camera rigid body values specified in
        with_aos_dof, but not other degrees of freedom.

        Parameters
        ----------
        dof : ndarray (5,), optional
            0,1,2 are camera z,x,y in micron
            3,4 are camera rot around x, y in arcsec
        dx, dy, dz : float
            camera displacement in micron
        rx, ry : galsim.Angle
            camera rotation around x, y

        Returns
        -------
        ret : SSTBuilder
            New builder with specified camera rigid body DOF.
        """
        ret = copy(self)
        if (
            any([x is not None for x in [dx, dy, dz, rx, ry]])
            and dof is not None
        ):
            raise ValueError("Cannot specify both dof and dx,dy,dz,rx,ry")
        if dof is not None:
            ret.dof[5:10] = dof
        else:
            if dz is not None:
                ret.dof[5] = dz
            if dx is not None:
                ret.dof[6] = dx
            if dy is not None:
                ret.dof[7] = dy
            if rx is not None:
                ret.dof[8] = rx.deg*3600
            if ry is not None:
                ret.dof[9] = ry.deg*3600
        return ret

    @attach_attr(
        _req_params={
            "dof":None,
        },
    )
    def with_m1m3_bend(self, dof):
        """Return new SSTBuilder that includes specified M1M3 bending modes

        Parameters
        ----------
        dof : ndarray (len(use_m1m3_modes),)
            M1M3 bending modes in micron

        Returns
        -------
        ret : SSTBuilder
            New builder with specified M1M3 bending modes.
        """
        ret = copy(self)
        ret.dof[self.m1m3_dof_indices] = dof
        return ret

    @attach_attr(
        _req_params={
            "dof":None,
        },
    )
    def with_m2_bend(self, dof):
        """Return new SSTBuilder that includes specified M2 bending modes

        Parameters
        ----------
        dof : ndarray (len(use_m2_modes),)
            M2 bending modes in micron

        Returns
        -------
        ret : SSTBuilder
            New builder with specified M2 bending modes.
        """
        ret = copy(self)
        ret.dof[self.m2_dof_indices] = dof
        return ret

    @attach_attr(
        _req_params={
            "zk":None,
            "eps":float,
        },
    )
    def with_extra_zk(self, zk, eps):
        """Return new SSTBuilder that includes specified constant Zernike phase
        screen at entrance pupil.

        Parameters
        ----------
        zk : ndarray
            Zernike coefficients of phases to add at entrance pupil in meters.
        eps : float
            Use annular Zernikes with this fractional obscuration.

        Returns
        -------
        ret : SSTBuilder
            New builder with specified extra Zernike phases.
        """
        ret = copy(self)
        ret.extra_zk = zk
        ret.extra_zk_eps = eps
        return ret

    @attach_attr(
        _req_params={
            "forces":None,
        },
    )
    def with_m1m3_extra_forces(self, forces):
        """Return new SSTBuilder that includes specified M1M3 extra forces

        Parameters
        ----------
        forces : ndarray (256,)
            M1M3 extra forces in N

        Returns
        -------
        ret : SSTBuilder
            New builder with specified M1M3 extra forces.
        """
        # Todo: should be able to index these by actuator ID too.
        ret = copy(self)
        ret.m1m3_extra_forces = forces
        return ret

    def build(self):
        optic = self.fiducial
        optic = self._apply_phase(optic)
        optic = self._apply_rigid_body_perturbations(optic)
        optic = self._apply_M1M3_surface_perturbations(optic)
        optic = self._apply_M2_surface_perturbations(optic)
        optic = self._apply_camera_surface_perturbations(optic)
        return optic

    def _apply_phase(self, optic):
        if self.extra_zk is None:
            return optic
        optic = optic.withInsertedOptic(
            before="M1",
            item=batoid.OPDScreen(
                name='Screen',
                surface=batoid.Plane(),
                screen=batoid.Zernike(
                    self.extra_zk,
                    R_outer=4.18,
                    R_inner=self.extra_zk_eps*4.18
                ),
                coordSys=optic.stopSurface.coordSys,
                obscuration=optic['M1'].obscuration,
            )
        )
        return optic

    def _apply_rigid_body_perturbations(self, optic):
        dof = self.dof
        if np.any(dof[0:3]):
            optic = optic.withGloballyShiftedOptic(
                "M2",
                np.array([-dof[1], dof[2], -dof[0]])*1e-6
            )

        if np.any(dof[3:5]):
            rx = batoid.RotX(np.deg2rad(-dof[3]/3600))
            ry = batoid.RotY(np.deg2rad(dof[4]/3600))
            optic = optic.withLocallyRotatedOptic(
                "M2",
                rx @ ry
            )

        if np.any(dof[5:8]):
            optic = optic.withGloballyShiftedOptic(
                self.cam_name,
                np.array([-dof[6], dof[7], -dof[5]])*1e-6
            )

        if np.any(dof[8:10]):
            rx = batoid.RotX(np.deg2rad(-dof[8]/3600))
            ry = batoid.RotY(np.deg2rad(dof[9]/3600))
            optic = optic.withLocallyRotatedOptic(
                self.cam_name,
                rx @ ry
            )

        return optic

    def _apply_M1M3_surface_perturbations(self, optic):
        dof = self.dof
        # Collect gravity/temperature perturbations
        # be sure to make a copy here!
        m1m3_fea = np.array(m1m3_gravity(
            self.fea_dir, self.fiducial, self.m1m3_zenith
        ))
        m1m3_fea += m1m3_temperature(
            self.fea_dir,
            self.m1m3_TBulk,
            self.m1m3_TxGrad,
            self.m1m3_TyGrad,
            self.m1m3_TzGrad,
            self.m1m3_TrGrad
        )

        m1m3_forces = np.array(m1m3_lut(
            self.fea_dir,
            self.m1m3_lut_zenith,
            self.m1m3_lut_error,
            self.m1m3_lut_seed
        ))
        m1m3_forces += self.m1m3_extra_forces

        m1m3_fea += m1m3_force_to_surface(
            self.fea_dir,
            m1m3_forces
        )

        bx, by, idx1, idx3 = m1m3_fea_nodes(self.fea_dir)
        if np.any(m1m3_fea):
            # Decompose into independent annular Zernikes+grid on M1 and M3
            m1_fea = m1m3_fea[idx1]
            m3_fea = m1m3_fea[idx3]

            zBasis1 = galsim.zernike.zernikeBasis(
                28, bx[idx1], by[idx1], R_outer=4.18, R_inner=2.558
            )
            zBasis3 = galsim.zernike.zernikeBasis(
                28, bx[idx3], by[idx3], R_outer=2.508, R_inner=0.55
            )
            m1_zk, *_ = np.linalg.lstsq(zBasis1.T, m1_fea, rcond=None)
            m3_zk, *_ = np.linalg.lstsq(zBasis3.T, m3_fea, rcond=None)
            m1_zk = galsim.zernike.Zernike(m1_zk, R_outer=4.18, R_inner=2.558)
            m3_zk = galsim.zernike.Zernike(m3_zk, R_outer=2.508, R_inner=0.55)
            m1_fea -= m1_zk(bx[idx1], by[idx1])
            m3_fea -= m3_zk(bx[idx3], by[idx3])

            m1_grid = _node_to_grid(
                bx[idx1], by[idx1], m1_fea, m1m3_grid_xy(self.bend_dir)[0]
            )
            m3_grid = _node_to_grid(
                bx[idx3], by[idx3], m3_fea, m1m3_grid_xy(self.bend_dir)[1]
            )
        else:
            m1_nx = m1m3_grid_xy(self.bend_dir)[0].shape[1]
            m3_nx = m1m3_grid_xy(self.bend_dir)[1].shape[1]
            m1_grid = np.zeros((4, m1_nx, m1_nx))
            m3_grid = np.zeros((4, m3_nx, m3_nx))
            m1_zk = galsim.zernike.Zernike(
                np.zeros(29), R_outer=4.18, R_inner=2.558
            )
            m3_zk = galsim.zernike.Zernike(
                np.zeros(29), R_outer=2.508, R_inner=0.55
            )

        # Fold in M1M3 bending modes
        if np.any(dof[self.m1m3_dof_indices]):
            bend1 = realize_bend(
                self.bend_dir,
                tuple(dof[self.m1m3_dof_indices]),
                tuple(self.use_m1m3_modes),
                "M1"
            )
            bend3 = realize_bend(
                self.bend_dir,
                tuple(dof[self.m1m3_dof_indices]),
                tuple(self.use_m1m3_modes),
                "M3"
            )

            m1_zk += transform_zernike(bend1.zk, R_outer=4.18, R_inner=2.558)
            m3_zk += transform_zernike(bend3.zk, R_outer=2.508, R_inner=0.55)

            m1_grid += bend1.grid
            m3_grid += bend3.grid

        # Apply to M1
        components = []
        if np.any(m1_zk.coef):
            components.append(
                batoid.Zernike(
                    m1_zk.coef, m1_zk.R_outer, m1_zk.R_inner
                )
            )
        if np.any(m1_grid):
            components.append(
                batoid.Bicubic(
                    *m1m3_grid_xy(self.bend_dir)[0],
                    *m1_grid,
                    nanpolicy='zero'
                )
            )
        if components:
            if len(components) >= 2:
                perturbation = batoid.Sum(components)
            else:
                perturbation = components[0]
            optic = optic.withPerturbedSurface(
                'M1',
                perturbation
            )

        # Apply to M3
        components = []
        if np.any(m3_zk.coef):
            components.append(
                batoid.Zernike(
                    m3_zk.coef, m3_zk.R_outer, m3_zk.R_inner
                )
            )
        if np.any(m3_grid):
            components.append(
                batoid.Bicubic(
                    *m1m3_grid_xy(self.bend_dir)[1],
                    *m3_grid,
                    nanpolicy='zero'
                )
            )
        if components:
            if len(components) >= 2:
                perturbation = batoid.Sum(components)
            else:
                perturbation = components[0]
            optic = optic.withPerturbedSurface(
                'M3',
                perturbation
            )
        return optic

    def _apply_M2_surface_perturbations(self, optic):
        dof = self.dof
        # Collect gravity/temperature perturbations
        # be sure to make a copy here!
        m2_fea = np.array(m2_gravity(
            self.fea_dir, self.m2_zenith
        ))
        m2_fea += m2_temperature(
            self.fea_dir,
            self.m2_TzGrad,
            self.m2_TrGrad
        )
        # m2_fea += m2_lut(
        #     self.fea_dir,
        #     self.m2_lut_zenith,
        # )

        bx, by = m2_fea_nodes(self.fea_dir)
        if np.any(m2_fea):
            # Decompose into annular Zernikes+grid on M2
            zBasis2 = galsim.zernike.zernikeBasis(
                28, bx, by, R_outer=1.71, R_inner=0.9
            )
            m2_zk, *_ = np.linalg.lstsq(zBasis2.T, m2_fea, rcond=None)
            m2_zk = galsim.zernike.Zernike(m2_zk, R_outer=1.71, R_inner=0.9)
            m2_fea -= m2_zk(bx, by)

            m2_grid = _node_to_grid(
                bx, by, m2_fea, m2_grid_xy(self.bend_dir)
            )
        else:
            m2_nx = m2_grid_xy(self.bend_dir).shape[1]
            m2_grid = np.zeros((4, m2_nx, m2_nx))
            m2_zk = galsim.zernike.Zernike(
                np.zeros(29), R_outer=1.71, R_inner=0.9
            )

        # Fold in M2 bending modes
        if np.any(dof[self.m2_dof_indices]):
            bend2 = realize_bend(
                self.bend_dir,
                tuple(dof[self.m2_dof_indices]),
                tuple(self.use_m2_modes),
                "M2"
            )
            m2_zk += transform_zernike(bend2.zk, R_outer=1.71, R_inner=0.9)
            m2_grid += bend2.grid

        # Apply to M2
        components = []
        if np.any(m2_zk.coef):
            components.append(
                batoid.Zernike(
                    m2_zk.coef, m2_zk.R_outer, m2_zk.R_inner
                )
            )
        if np.any(m2_grid):
            components.append(
                batoid.Bicubic(
                    *m2_grid_xy(self.bend_dir),
                    *m2_grid,
                    nanpolicy='zero'
                )
            )
        if components:
            if len(components) >= 2:
                perturbation = batoid.Sum(components)
            else:
                perturbation = components[0]
            optic = optic.withPerturbedSurface(
                'M2',
                perturbation
            )
        return optic

    def _apply_camera_surface_perturbations(self, optic):
        if self.camera_zenith is None:
            zen = None
            rot = None
        else:
            zen = self.camera_zenith
            rot = self.camera_rotation
        TBulk = self.camera_TBulk

        for tname, bname, radius in [
            ('L1S1', 'L1_entrance', 0.775),
            ('L1S2', 'L1_exit', 0.775),
            ('L2S1', 'L2_entrance', 0.551),
            ('L2S2', 'L2_exit', 0.551),
            ('L3S1', 'L3_entrance', 0.361),
            ('L3S2', 'L3_exit', 0.361)
        ]:
            data = _fits_cache(self.fea_dir, tname+"zer.fits.gz")
            zk = np.zeros(28)
            if zen is not None:
                zk += data[0, 3:] * (np.cos(zen) - 1)
                zk += (
                    data[1, 3:] * np.cos(rot) +
                    data[2, 3:] * np.sin(rot)
                ) * np.sin(zen)

            if TBulk is not None:
                TBulk1 = np.clip(
                    TBulk,
                    np.min(data[3:, 2])+0.001,
                    np.max(data[3:, 2])-0.001
                )
                fidx = np.interp(
                    TBulk1, data[3:, 2], np.arange(len(data[3:, 2]))
                )+3
                idx = int(np.floor(fidx))
                whi = fidx - idx
                wlo = 1 - whi
                zk += wlo * data[idx, 3:] + whi * data[idx+1, 3:]

                # subtract reference temperature zk (0 deg C is idx=5)
                zk -= data[5, 3:]

            # remap Andy -> Noll Zernike indices
            zIdxMapping = [
                1, 3, 2, 5, 4, 6, 8, 9, 7, 10, 13, 14, 12, 15, 11, 19, 18, 20,
                17, 21, 16, 25, 24, 26, 23, 27, 22, 28
            ]
            zk = zk[[x - 1 for x in zIdxMapping]]
            zk *= 1e-3  # mm -> m
            # tsph -> batoid 0-index offset
            zk = np.concatenate([[0], zk])

            # Now need to flip x and z for Zemax -> batoid
            for j in range(1, 29):
                n, m = galsim.zernike.noll_to_zern(j)
                if (n+(m>=0)) % 2 == 0:  # antisymmetric in x
                    zk[j] *= -1
            zk = -zk

            if np.any(zk):
                optic = optic.withPerturbedSurface(
                    bname,
                    batoid.Zernike(zk, R_outer=radius)
                )

        return optic



from pathlib import Path
from subprocess import run


zenodo_dois = {
    "bend_legacy": 8383823,
    "fea_legacy": 8384326,
    "bend": 8384775,
    "bend_ptt": 8384781,
}


def download_rubin_data(args):
    DOI = zenodo_dois.get(args.dataset, None)
    if DOI is None:
        raise ValueError(f"Unknown dataset {args.dataset}")
    cmd = f"zenodo_get {DOI}"
    if not args.outdir:
        outdir = Path(__file__).parent / args.dataset
    else:
        outdir = Path(args.outdir)
    cmd += f" -o {outdir}"
    run(cmd, shell=True)


if __name__ == "__main__":
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument(
        "dataset",
        type=str,
        choices=zenodo_dois.keys()
    )
    parser.add_argument(
        "--outdir",
        type=str,
        default=None
    )
    args = parser.parse_args()
    download_rubin_data(args)
